#!/usr/bin/python
# -*- coding: utf-8 -*-


from string import Template
import pandas as pd
import numpy as np
import re
import os
import shutil
import pandas as pd
import itertools
from datetime import datetime

from .utils import *
from .extract_fields import *
# from ..CONFIG import *



# """## Create views/raw/arrays view

# This is the view file that includes all first level arrays and structs of the main table
# """



def create_nested_dimension(field, add_links_dimensions = True, verbose=False):
  assert field.modus in ["ARRAY", "STRUCT"]


  if field.modus == "ARRAY":
    dimension_lml = f"""
    {make_header(field, "Links-dimensions for array", True)}
    dimension: {field.nested_dimension_name} {{
      hidden: yes
      description: "{field.description}"
      sql: ${{TABLE}}.{field.unnested_sql} ;;
    }}"""

    if add_links_dimensions:
      dimension_lml = dimension_lml + f"""
      dimension: links_{field.nested_dimension_name} {{
        type: number
        hidden:  no
        label: "Nr of {field.label_plural}"
        description: "Non-unique Nr of {field.label_plural}: {field.description}"
        group_label: "Links count"
        sql: ARRAY_LENGTH(${{{field.nested_dimension_name}}}) ;;
      }}
      dimension: has_{field.nested_dimension_name} {{
        description: "Yes if {get_global_configuration()['pretty_looker_name']} has any {field.label_plural}"
        type: yesno
        group_label: "Links check"
        sql: ${{links_{field.nested_dimension_name}}} > 0 ;;
      }}
      """

    if field.nested_dimension_name == "active_years":
      dimension_lml = dimension_lml +"""
    dimension: active_years_string {
      hidden: no
      label: "Active Years (combined)"
      sql: ( SELECT STRING_AGG(CAST(id AS STRING), ',')  FROM UNNEST(${active_years}) id);;
      group_label: "Dates"
    }
    """

  elif field.modus == "STRUCT":
    if verbose: printDebug(field.field_path)
    dimension_lml = f"""
    {make_header(field, "has_dimension for struct", True)}

    dimension: has_{field.nested_dimension_name} {{
      description: "Yes if {get_global_configuration()['pretty_looker_name']} has {field.label}"
      type: yesno
      group_label: "Links check"
      sql: ${{TABLE}}.{field.unnested_sql} is not null ;;
    }}
    """
  else:
    assert 1==1 # FIELDS cannot be used here

  return dimension_lml


#
# Create all the dimensions of the raw/array view
#
def create_all_raw_array_dimensions(looker_metadata, verbose=False):
  top_level_array_struct_fields = looker_metadata[(looker_metadata.nested_view_name == get_global_configuration()['looker_name']) & (looker_metadata.modus.isin(["ARRAY", "STRUCT"]))]
  if verbose:
    printDebug("Generating joins for the top level ", len(top_level_array_struct_fields), " arrays and struct found in the table")
    printDebug(top_level_array_struct_fields)
  return "".join(list(top_level_array_struct_fields.apply(create_nested_dimension, axis=1)))



def write_raw_array_view(output_path, looker_metadata):

    raw_array_view_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: BASE VIEWS - GBQ ARRAY FIELDS
    #
    # - represents GBQ fields 1:1
    # - array field is normally a hidden dimension (used in nested JOINs)
    # - extra dimensions for each array: links count
    # - optional dimensions for each array: HasValue Yes/no (shortcut for quick filtering)
    #
    ##############





    view: arrays_{get_global_configuration()['looker_name']} {{

    sql_table_name: `dimensions-ai.data_analytics.{get_global_configuration()['looker_name']}`
    ;;

    {create_all_raw_array_dimensions(looker_metadata)}

    }}
    """

    # printDebug(raw_array_view_lml)
    f = open(output_path+"/views/raw/array."+get_global_configuration()['looker_name']+".view.lkml", "w")
    f.write(raw_array_view_lml)
    f.close()






# """## Create views/raw/fields view

# This is the view file that includes all first level simple fields of the main table
# """

def create_unnested_dimension(field):
  # if not((field.modus == "FIELD") or (field.modus == "ARRAY" and field.unnested_dimension_name != "")):
  #   display(field)
  #   assert 1==0
  assert (field.modus == "FIELD") or (field.modus == "ARRAY" and 
                                      field.unnested_dimension_name != "" and
                                      not field.data_type.startswith('ARRAY<STRUCT<'))

  #Default values of some sections
  dimension_type = "dimension"
  dimension_name = field.unnested_dimension_name
  header = ""
  extras = ""
  group_label    = field.group_label
  label = field.label

  if field.modus == "FIELD":
    data_type = field.data_type   
    sql = f"${{TABLE}}.{field.unnested_sql}"
  else: # ARRAY
    data_type = field.data_type[6:-1]
    sql = f"${{TABLE}}"  

  if data_type in ["INT64", "FLOAT64", "BIGNUMERIC", "NUMERIC"]: 
    looker_type = "number"
  elif data_type == "BOOL":
    looker_type = "yesno"
  elif (data_type in ["DATE", "TIMESTAMP", "TIME", "DATETIME"]) or field.column_name.endswith("date") or field.column_name.startswith("date_") :
    looker_type = "time"
    dimension_type = "dimension_group"
    extras = """
    timeframes: [
      raw,
      date,
      week,
      month,
      quarter,
      year
    ]
    convert_tz: no
    datatype: timestamp"""

    if field.column_name in ["date_inserted", "date_imported_gbq"]:
      group_label= "Dates - Internal"
    else:
      group_label= "Dates"
    if data_type == "STRING":
      sql = f"CAST({sql} AS DATE)"
 
  elif data_type == "STRING":
    looker_type = "string"
    if "country" in dimension_name or "countries" in dimension_name or "jurisdiction" in dimension_name:
      extras = extras+"""
    map_layer_name: countries"""
  else:
    printDebug("Unknown type:", data_type)
    printDebug(field)
    printDebug(f"""
    #
    # ££££££££
    # Cannot generate code for {field.field_path} because of
    # type {data_type} 
    """)
    assert 1 == 2

  if field.field_path == "id":
    group_label= "Identifiers"
    extras = f"""
    primary_key: yes"""
    label = " Dimensions ID" # --- LEADING SPACE ---
  
  if dimension_name.endswith("_id"):
    group_label= "Identifiers"

  if dimension_name.endswith("year") and looker_type=="number":
    group_label = "Dates"
    label = label +" (number)"
    extras= """
    value_format: "0000" """

  # For some attributes we may need special casts see at the beginning
  sql = get_global_configuration()['special_casts'].get(dimension_name, sql)

  dimension_lml =f"""{make_header(field, "", False)}
  {dimension_type}: {dimension_name} {{
    type:         {looker_type}
    label:        "{label}"
    description:  "{field.description}"{extras}
    group_label:  "{group_label}"
    sql: {sql};;{get_link(field)}
  }}
  """

  # Add a full name for good measure 
  if dimension_name == "last_name":
    dimension_lml = dimension_lml + f"""
  dimension: full_name {{
    type:         string
    label:        "Full Name"
    group_label:  "{group_label}"  #£££ CHECK THIS!
    sql: COALESCE(CONCAT(${{first_name}}, ' ', ${{last_name}}), ${{last_name}}, ${{first_name}});;
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/entities/publication/author/{{{{ researcher_id._value }}}}" #£££ check if correct value
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}
  }}"""

  return dimension_lml
  
#
# Create all the dimensions of the raw/field view
#
def create_all_raw_fields_dimensions(looker_metadata, verbose=False):
  fields = looker_metadata[(looker_metadata.unnested_view_name == get_global_configuration()['looker_name']) & (looker_metadata.modus.isin(["FIELD"]))]
  if verbose:
    printDebug("Generating joins for the top level ", len(fields), " simple fields found in the table")
    printDebug(fields)
  return "".join(list(fields.apply(create_unnested_dimension, axis=1)))



def write_raw_fields_view(output_path, looker_metadata):

    raw_fields_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: BASE VIEWS - MISC FIELDS
    #
    # - represents GBQ fields 1:1
    #
    # https://docs.dimensions.ai/bigquery/datasource-{get_global_configuration()['looker_name']}.html
    ##############


    view: fields_{get_global_configuration()['looker_name']} {{

    sql_table_name: `dimensions-ai.data_analytics.{get_global_configuration()['looker_name']}`
    ;;

    {create_all_raw_fields_dimensions(looker_metadata)}

    }}
    """
    f = open(output_path+"/views/raw/fields."+get_global_configuration()['looker_name']+".view.lkml", "w")
    f.write(raw_fields_lml)
    f.close()






# """## Create views/unnested/array.nested view

# This is the view file that includes all the unnesting of the arrays in the main table
# """



# Create a certain dimension in a particular view
# for a nested array there are two options:
# 1. The unnested array is a hidden dimension in the parent array
# 2. If the array is ARRAY<SIMPLE TYPE> i.e. not ARRAY <STRUCT> then this field is a dimension of its own unnested view
#.   (Example is eg. publications.research_org)
def create_array_unnested_view_dimension(field):
#   display(field)
  if field.modus == "FIELD":
    return create_unnested_dimension(field)
  elif field.modus == "ARRAY":
    return create_nested_dimension(field, False)
  elif field.modus == "STRUCT":
    #do nothing
    return ""
  else:
    assert 1 == 2 # unknown modus

def create_array_unnested_view(field, looker_metadata):
  pat = field.field_path+"\.[a-z_]+$"
  sub_fields = looker_metadata[looker_metadata.field_path.str.fullmatch(pat)]

  if len(sub_fields) > 0:
    # This is an ARRAY of STRUCT: then all fields inside appear as separate fields
    assert field.unnested_dimension_name == ""
    dimensions_lml = "".join(list(sub_fields.apply(create_array_unnested_view_dimension, axis=1)))
  else:
    # This is an ARRAY<STRING> (or another simple type): the field is the "same" as the array
    assert field.unnested_dimension_name != ""
    dimensions_lml = create_unnested_dimension(field)


  if field.field_path in get_global_configuration()['fields_w_offset_joins'] :
    dimensions_lml = f"""
    dimension: {field.field_path}_key {{
      type: string
      hidden: yes
      primary_key: yes

      # TIP: in the explore join, an OFFSET is created and given the alias authors_offset
      # https://cloud.google.com/bigquery/docs/reference/standard-sql/query-syntax#unnest_operator
      sql: CONCAT({field.unnested_sql}_offset, ${{{get_global_configuration()['looker_name']}.id}}) ;;
    }}"""+dimensions_lml

  if field.field_path == "concepts":
    dimensions_lml = dimensions_lml + """
    dimension: is_0_5_plus {
      description: "Yes if concept has a relevance score major than 0.5"
      label: "Has Relevance 0.5+"
      hidden: no
      type: yesno
      group_label: "Concepts"
      sql: ${relevance} > 0.5 ;;
    }

    dimension: is_0_6_plus {
      description: "Yes if concept has a relevance score major than 0.6"
      label: "Has Relevance 0.6+"
      hidden: no
      type: yesno
      group_label: "Concepts"
      sql: ${relevance} > 0.6 ;;
    }
    """

  view_lml = f"""
{make_header(field, "Unnested view for ")}
view: {field.unnested_view_name} {{
  {dimensions_lml}
}}
  """

  return view_lml

def create_all_array_unnested_views(looker_metadata, verbose=False):
  fields = looker_metadata[(looker_metadata.unnested_view_name != get_global_configuration()['looker_name']) & (looker_metadata.modus=="ARRAY")]
  if verbose:
    printDebug("Generating joins for the not top level ", len(fields), " arrays  found in the table")
    printDebug(fields)
  return "".join(list(fields.apply(create_array_unnested_view, axis=1, looker_metadata=looker_metadata)))


def write_unested_array_view(output_path, looker_metadata):

    unnested_array_nested_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: NESTED FIELDS VIEWS - MISCELLANEOUS
    #
    # - represents GBQ nested fields by adding a view for each of them
    #
    # https://docs.dimensions.ai/bigquery/datasource-{get_global_configuration()['looker_name']}.html
    ##############

    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/views/base.view.lkml"
    {create_all_array_unnested_views(looker_metadata)}
    """

    f = open(output_path+"/views/unnested/array.nested.view.lkml", "w")
    f.write(unnested_array_nested_lml)
    f.close()





# """## Create views/base
# This mainly includes:
# - Measures attached to "links" aka arrays
# - Measures attached to number-type of attributes
# """

#
# Build simple measures for number-type
#
def create_base_field_measure(field):
  dimension_ml = f"""{make_header(field, "Measures for ")}
  measure: total_{field.unnested_dimension_name} {{
    type: sum
    label: "Sum {field.label_plural}"
    sql:  ${{{field.unnested_dimension_name}}} ;;
    group_label: "{field.group_label}"
  }}
  measure: avg_{field.unnested_dimension_name} {{
    type: average
    label: "Average {field.label_plural}"
    sql:  ${{{field.unnested_dimension_name}}} ;;
    group_label: "{field.group_label}"
  }}
  """
  return dimension_ml

def create_all_base_field_measures(looker_metadata, verbose=False):
  fields = looker_metadata[looker_metadata.data_type.isin(['INT64', 'FLOAT64', "BOOL"])&
          (~looker_metadata.field_path.str.contains("year")) &
          (looker_metadata.field_path != "concepts.relevance") &
          (looker_metadata.unnested_view_name == get_global_configuration()['looker_name'])]
  if verbose:
    printDebug("Generating joins for the  top level ", len(fields), " fields  found in the table")
    # display(fields)
  return "".join(list(fields.apply(create_base_field_measure, axis=1)))


#
# Create measure for nested attributes i.e. "links"
#
def create_array_measure(field):

  # Not the whole shebang for active years
  if field.field_path == "active_years":
    return f"""
  {make_header(field, "Measures for ")}
  measure: avg_{field.nested_dimension_name}  {{
    type: average
    value_format_name: decimal_2
    label: "Average {field.label}"
    description: "When used on a single {get_global_configuration()['pretty_looker_name']}, returns the average number of active years"
    group_label: "{field.group_label}"
    sql: ${{links_{field.nested_dimension_name}}} ;;
    drill_fields: [detail*]
  }}    
    """

  measure_lml = make_header(field, "Measures for ", True)
  if field.modus == "ARRAY":
    measure_lml = measure_lml+f"""
  measure: total_links_{field.nested_dimension_name}   {{
    type: sum
    label: "Total {field.label_plural}"
    description: "Non unique total number of all {field.label_plural}"
    group_label: "{field.group_label}"
    sql: ${{links_{field.nested_dimension_name}}} ;;
    drill_fields: [detail*]
  }}
  measure: avg_links_{field.nested_dimension_name}  {{
    type: average
    value_format_name: decimal_2
    label: "Average {field.label}"
    description: "Non unique: average  {field.label} links per {get_global_configuration()['pretty_looker_name']}"
    group_label: "{field.group_label}"
    sql: ${{links_{field.nested_dimension_name}}} ;;
    drill_fields: [detail*]
  }}"""
  else:
    measure_lml = ""

  measure_lml = measure_lml + f"""
  measure: count_{get_global_configuration()['looker_name']}_with_{field.nested_dimension_name}  {{
    type: count
    label: "Unique {get_global_configuration()['pretty_looker_name']} with {field.label_plural}"
    group_label: "{field.group_label}"
    filters: [has_{field.nested_dimension_name} : "Yes"]
    drill_fields: [detail*]
  }}
  measure: percent_{get_global_configuration()['looker_name']}_with_{field.nested_dimension_name}  {{
    type: number
    value_format_name: percent_2
    label: "Percent {get_global_configuration()['pretty_looker_name']} with a {field.label_plural}"
    group_label: "{field.group_label}"
    sql:  ${{count_{get_global_configuration()['looker_name']}_with_{field.nested_dimension_name}}} / ${{count}} ;;
    drill_fields: [detail*]
  }}
  """
  return measure_lml

def create_all_link_measures(looker_metadata, verbose=False):
  fields = looker_metadata[(looker_metadata.nested_view_name == get_global_configuration()['looker_name']) & (looker_metadata.modus.isin(["ARRAY", "STRUCT"]))]
  if verbose:
    printDebug("Generating joins for the top level ", len(fields), " arrays  found in the table")
    # display(fields)
  return "".join(list(fields.apply(create_array_measure, axis=1)))


def write_base_view(output_path, looker_metadata):


    base_view_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']} BASE VIEW
    #
    # - raw fields as they appear in GBQ
    # - basic measures
    # - no 'unnested' views dependencies
    #
    ##############

    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/views/raw/*.view"


    view: {get_global_configuration()['looker_name']}_base {{

    extends: [arrays_{get_global_configuration()['looker_name']}, fields_{get_global_configuration()['looker_name']}]
    sql_table_name: `dimensions-ai.data_analytics.{get_global_configuration()['looker_name']}` ;;

    label: "{get_global_configuration()['pretty_looker_name']}"


    ##############
    ##############
    ##
    # FIELD SETS FOR DRILLING
    ##
    ##############
    ##############



    # default
    drill_fields: [id]


    # ----- Sets of fields for drilling in ------
    set: detail {{
        fields: [{
            ", ".join(looker_metadata[
            (looker_metadata.unnested_view_name == get_global_configuration()['looker_name']) & 
            (looker_metadata.modus == "FIELD") ].unnested_dimension_name.head(4))
            }]
    }}



    ##############
    ##############
    ##
    # MEASURES derived from RAW fields
    ##
    ##############
    ##############

    measure: count {{
        type: count
        label: "Unique {get_global_configuration()['pretty_looker_name']}"
        group_label: " {get_global_configuration()['pretty_looker_name']}"
        drill_fields: [detail*]
    }}
    

    #######################
    #  MEASURES FOR NUMERIC FIELDS
    #
    {create_all_base_field_measures(looker_metadata)}


    #######################
    #  MEASURES FOR ARRAYS/LINKS
    #
    {create_all_link_measures(looker_metadata)}

    }}
    """

    f = open(output_path+"/views/base.view.lkml", "w")
    f.write(base_view_lml)
    f.close()







# """## Create views/unnested

# This view includes all measures which are counts distinct aka unique counts AFTER unnesting. For this we list all ARRAYs of a simple data type like INT64 or STRING and create counts distinct measures.

# """

def create_unnested_measure(field):
  measure_lml = f"""
  {make_header(field, "Measures for ")}
  measure: count_{field.field_path.replace(".","_")} {{
    type: count_distinct
    label: "Unique {field.label_plural}"
    description: "Number of unique {field.label_plural}"
    group_label: "{field.group_label}"
    sql: ${{{field.unnested_view_name}.{field.unnested_dimension_name}}} ;;
  }}
  """
  return measure_lml

def create_all_unnested_measures(looker_metadata, verbose=False):
  fields = looker_metadata[looker_metadata.data_type.str.fullmatch(r'ARRAY<[A-Z0-9]*?>')]
  if verbose:
    printDebug("Generating joins for the  ", len(fields), " fields  found in the table")
    # display(fields)
  return "".join(list(fields.apply(create_unnested_measure, axis=1)))



def write_unnested_view(output_path, looker_metadata):

    unnested_views_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: UNNESTED VIEW
    #
    # - extend BASE by adding unnesting logic
    # - additional measures
    #
    # NOTE: when importing this view into an Explore, you need all
    # the unnesting JOINs for the symbols to resolve
    ##############

    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/views/base.view"
    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/views/unnested/*.view"


    view: {get_global_configuration()['looker_name']}_unnested {{

    extends: [{get_global_configuration()['looker_name']}_base]


    ##############
    ##############
    ##
    # MEASURES THAT REQUIRE UNNESTED DATA
    ##
    ##############
    ##############

    {create_all_unnested_measures(looker_metadata)}

    }}
    """


    f = open(output_path+"/views/unnested.view.lkml", "w")
    f.write(unnested_views_lml)
    f.close()


# !/usr/bin/env python
# -*- coding: utf-8 -*-


from string import Template
import click
import pandas as pd
import numpy as np
import re
import os
import shutil
import pandas as pd
import itertools
from datetime import datetime


CONFIGURATION = {}

def get_global_configuration():
    global CONFIGURATION
    return CONFIGURATION






def gen_folders(user_provided_name, default_dir):
  """ """
  if not os.path.exists(default_dir):
      os.mkdir(default_dir)

  output_path = os.path.join(default_dir, user_provided_name)
  if os.path.exists(output_path):
      shutil.rmtree(output_path)

  os.mkdir(output_path)
  os.mkdir(output_path+"/explores")
  os.mkdir(output_path+"/views")
  os.mkdir(output_path+"/views/raw")
  os.mkdir(output_path+"/views/unnested")

  with open(output_path+'/README.MD', 'w') as f:
      f.write(gen_readme())

  return output_path


def gen_readme():
  """Simple README.MD file contents"""

  now = datetime.now() # current date and time
  date_time = now.strftime("%m/%d/%Y, %H:%M:%S")
  text = f"""Auto-generated LookML project
  
  Date: {date_time}
  """
  return text





def make_header(field, template, include_struct=False):
  header = ""
  if field.top_header != "":
    if not(include_struct and field.field_path != field.column_name):
      header = f"""
  # ##############
  # ## {field.top_header}
  # #
  """

  if field.column_name != field.field_path:
    header = header+ f"""
    #
    ## {template} {field.field_path}
    # """
  return header







def printDebug(text, mystyle="", err=True, **kwargs):
    """Wrapper around click.secho() for printing in colors with various defaults.

    :kwargs = you can do printDebug("s", bold=True)

    2018-12-06: by default print to standard error stderr (err=True)
    https://click.palletsprojects.com/en/5.x/api/#click.echo
    This means that the output is ok with `less` and when piped to other commands (or files)

    Styling output:
    <http://click.pocoo.org/5/api/#click.style>
    Styles a text with ANSI styles and returns the new string. By default the styling is self contained which means that at the end of the string a reset code is issued. This can be prevented by passing reset=False.

    This works also with inner click styles eg

    ```python
    uri, title = "http://example.com", "My ontology"
    printDebug(click.style("[%d]" % 1, fg='blue') +
               click.style(uri + " ==> ", fg='black') +
               click.style(title, fg='red'))
    ```

    Or even with Colorama

    ```
    from colorama import Fore, Style

    printDebug(Fore.BLUE + Style.BRIGHT + "[%d]" % 1 + 
            Style.RESET_ALL + uri + " ==> " + Fore.RED + title + 
            Style.RESET_ALL)
    ```


    Examples:

    click.echo(click.style('Hello World!', fg='green'))
    click.echo(click.style('ATTENTION!', blink=True))
    click.echo(click.style('Some things', reverse=True, fg='cyan'))
    Supported color names:

    black (might be a gray)
    red
    green
    yellow (might be an orange)
    blue
    magenta
    cyan
    white (might be light gray)
    reset (reset the color code only)
    New in version 2.0.

    Parameters:
    text – the string to style with ansi codes.
    fg – if provided this will become the foreground color.
    bg – if provided this will become the background color.
    bold – if provided this will enable or disable bold mode.
    dim – if provided this will enable or disable dim mode. This is badly supported.
    underline – if provided this will enable or disable underline.
    blink – if provided this will enable or disable blinking.
    reverse – if provided this will enable or disable inverse rendering (foreground becomes background and the other way round).
    reset – by default a reset-all code is added at the end of the string which means that styles do not carry over. This can be disabled to compose styles.

    """

    if mystyle == "comment":
        click.secho(text, dim=True, err=err)
    elif mystyle == "important":
        click.secho(text, bold=True, err=err)
    elif mystyle == "normal":
        click.secho(text, reset=True, err=err)
    elif mystyle == "red" or mystyle == "error":
        click.secho(text, fg='red', err=err)
    elif mystyle == "green":
        click.secho(text, fg='green', err=err)
    else:
        click.secho(text, err=err, **kwargs)




def printInfo(text, mystyle="", **kwargs):
    """Wrapper around printDebug for printing ALWAYS to stdout
    This means that the output can be grepped etc..
    NOTE this output will be picked up by pipes etc..

    Fixes https://github.com/lambdamusic/Ontospy/issues/76
    """
    printDebug(text, mystyle, False, **kwargs)






#################
# DIMENSIONS specific helpers
#################


# This may require adaption depending on what table you got
def get_link(field):
  if field.field_path == "id":
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/{get_global_configuration()['gbq_table_name'][:-1]}/{{{{ value  | url_encode | url_encode }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}"""

  elif field.field_path == "link_out":
    return f"""
    link: {{
      label: "Original URL"
      url: "{{{{ value  }}}}"
    }}"""

  elif field.unnested_dimension_name == "altmetric_id":
    return f"""
    link: {{
      label: "Altmetric"
      url: "https://www.altmetric.com/details/{{{{ value  }}}}"
      icon_url: "https://www.digital-science.com/wp-content/uploads/2020/11/Altmetric-no-padding.webp"
    }}"""

  elif field.unnested_dimension_name.endswith("grant_id"):
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/grant/{{{{ value  }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}"""

  elif field.unnested_dimension_name.endswith("publication_id"):
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/publication/{{{{ value  }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}"""

  elif field.unnested_dimension_name.endswith("grid_id") and "funder" in field.field_path:
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/{get_global_configuration()['gbq_table_name'][:-1]}/and_facet_funder={{{{ value  }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}"""
  
  elif field.unnested_dimension_name.endswith("grid_id"):
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/{get_global_configuration()['gbq_table_name'][:-1]}/and_facet_research_org={{{{ value  }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"
    }}"""   

  elif field.unnested_dimension_name == "researcher_id":
    return f"""
    link: {{
      label: "Dimensions"
      url: "https://app.dimensions.ai/details/entities/publication/author/{{{{ value  }}}}"
      icon_url: "https://www.dimensions.ai/wp-content/plugins/dimensions-blocks/assets/images/dimensions-hexagon-gradient.png"    
    }}"""
  else:
    return ""



#!/usr/bin/python
# -*- coding: utf-8 -*-


"""
simple test queries [for DEVELOPMENT  / not part of official tests]
"""

import click

from .. import *


@click.command()
@click.argument('test_number')
def quicktest_cli(test_number=1):

    test_number = int(test_number)

    if test_number == 1:
        return True




if __name__ == '__main__':
    quicktest_cli()



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit test stub
"""

from __future__ import print_function
import unittest, os, sys, click


class TestOne(unittest.TestCase):

    """
    Tests
    """

    click.secho("**TEST**", fg='red')

    def test_001(self):
        printDebug("Nothing to test yet")



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from .VERSION import __version__, VERSION
# make class available to global scope
from .gbqgen.main import run_gbq_extraction as from_gbq




# import lookmlhelper
# lookmlhelper.from_gbq()


# !/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import re
from textblob import Word  # making singulars and plurals

from .utils import *


# Get the tables with all fields and nesting for the GBQ Dimensions tables
def inspect_gbq(sql, verbose=False):
  CONFIGURATION = get_global_configuration()  
  if verbose:
    printDebug("==========\nQuerying:")
    printDebug(sql)
    printDebug("...")
  df = pd.io.gbq.read_gbq(sql, project_id = CONFIGURATION['gbq_project_id'])
  if verbose:
    printDebug("\t...loaded ", len(df), " records")
    # display(df)
  return df



# For convenience introduce the the mode
def get_gbq_mode(t, verbose=False):
  if t.startswith("ARRAY"):
    return "ARRAY"
  elif t.startswith("STRUCT"):
    return "STRUCT"
  else:
    return "FIELD"





# """### Create view and dimension names for all fields

# In the following we create the names of all views and dimensions for all attributes

# For arrays we need to construct 2 types of view/dimensions (e.g. patents: orange_book.ingredients)
# 1. UNNESTED: e.g. patents__orange_book__ingredients__unnested.ingredients
# 2. NESTED: patents.orange_book__ingredients

# There are a couple of complications:
# 1. We do not explicitely unnest structs - we simply reference them as struct.attribute
# 2. If an ARRAY is of a single variable (e.g. patents: researcher_ids) the attribute (researcher_ids) is part of the nested path names
# 3. We always add an unnest suffix to all unnested views and dimensions

# Furthermore we also create "nice" labels for each attribute both plural and singular. Users can define their own rules in the first section

# """


def get_parent_field_path(field_path):
    # Helper functions two split a field_path
    path_pat = re.compile(r"(.*?)\.?([^\.]+)$")
    return path_pat.search(field_path).group(1)


def get_array_parent_field_path(field_path, looker_metadata):
    parent_field_path = get_parent_field_path(field_path)
    if parent_field_path:
        parent_field = looker_metadata[looker_metadata.field_path == parent_field_path]
        assert len(parent_field) == 1  # paranoia
        parent_field = parent_field.iloc[0]

        if parent_field.modus == "STRUCT":
            return get_array_parent_field_path(parent_field.field_path, looker_metadata)
        else:
            return parent_field.field_path
    else:
        return ""


def get_attribute(field):
    parent_path_len = len(field.array_parent_field_path)
    if parent_path_len == 0:
        name = field.field_path
    else:
        name = field.field_path[parent_path_len + 1 :]
    return name.replace(".", "_")


def get_unnested_dimension_name(field):
    # a struct or an array of a struct does not represent a dimension and therefore the unnested dimension is empty
    # However ARRAY<STRING> is a repeat dimension and therefore it represents a dimension which we need to turn into a singular
    # FIELD of course are a dimension
    if field.data_type.startswith("ARRAY<STRUCT") or field.data_type.startswith(
        "STRUCT"
    ):
        dimension_name = ""
    elif field.data_type.startswith("ARRAY<"):
        # Often such an ARRAY is plural e.g. grid_ids ARRAY<STRING>
        # We need to singularise this. Hence we split by "_" and singularise the last part of it
        dimension_name = get_attribute(field)
        m = re.search(r"(.*?_?)([^\_]+)$", dimension_name)
        dimension_name = m.group(1) + Word(m.group(2)).singularize()
    else:
        dimension_name = get_attribute(field)

    return dimension_name


def get_unnested_view_name(field):
    if field.modus == "STRUCT":  # We do not explicitely reference or unnest structs
        return ""
    else:
        if field.modus == "ARRAY":
            path = field.field_path
        elif field.array_parent_field_path == "":
            return get_global_configuration()['looker_name']
        else:
            path = field.array_parent_field_path
        return get_global_configuration()['looker_name'] + "__" + path.replace(".", "__") + get_global_configuration()['unnest_suffix']


def get_nested_dimension_name(field):
    if field.modus == "ARRAY" or field.modus == "STRUCT":
        if field.array_parent_field_path:
            good_dimension_path = field.field_path[
                len(field.array_parent_field_path) + 1 :
            ]
        else:
            good_dimension_path = field.field_path
        return good_dimension_path.replace(".", "__")
    else:
        return ""


def get_nested_view_name(field):
    if field.modus == "ARRAY" or field.modus == "STRUCT":
        if field.array_parent_field_path:
            return (
                get_global_configuration()['looker_name']
                + "__"
                + field.array_parent_field_path.replace(".", "__")
                + get_global_configuration()['unnest_suffix']
            )
        else:
            return get_global_configuration()['looker_name']
    else:
        return ""


def get_label(field_path):
    label = field_path.replace(".", "-")
    label = label.lower()  # always true but better be safe

    # Replace all underscore with blank spaces
    label = label.replace("_", " ").strip()
    label = Word(label).singularize()
    if label.endswith("statu"):  # fix a problem with textblob
        label = label + "s"
    label = label.title()

    for special_label, special_new_label in get_global_configuration()['special_labels'].items():
        label = re.sub(
            r"\b" + special_label + r"\b", special_new_label, label, flags=re.IGNORECASE
        )

    return label


def make_label_plural(label):
    # pluralize does not work with capitalised nouns e.g. Country => Countrys
    # Hence a hack

    m = re.search(r"(.*?\s?)([^\s]+)$", label)
    last_word = m.group(2)

    # Last part is all capital letters e.g. ID or abbrev
    if last_word == last_word.upper():
        last_word = m.group(2) + "s"
    else:
        last_word = Word(last_word.lower()).pluralize().title()

    return m.group(1) + last_word


def get_unnested_sql(field):
    if field.array_parent_field_path == "":
        sql = field.field_path
    else:
        sql = field.field_path[len(field.array_parent_field_path) + 1 :]

    # for some weird reason full has to be put in paranthess
    return re.sub(r"\bfull\b", "`full`", sql, flags=re.IGNORECASE)


def get_group_label(field, looker_metadata):
    label = ""
    if field.unnested_view_name == get_global_configuration()['looker_name']:
        if field.modus == "FIELD":
            if "date" in field.field_path or "year" in field.field_path:
                label = "Dates"
            else:
                label = " Core Metadata"
    else:
        label = get_parent_field_path(field.field_path)
        if label:
            label = looker_metadata[looker_metadata.field_path == label].iloc[0].label

    if label == "":
        return field.label
    else:
        return label




# !/usr/bin/env python
#  -*- coding: UTF-8 -*-

__version__ = "0.1"  # LATEST? =>  https://pypi.org/project/XXXXX/
__copyright__ = "CopyRight (C) 2021 by Donald Duck"
__license__ = "MIT"
__author__ = "Donald Duck"
__author_email__ = ""

USAGE = "lookmlhelper [options]"
VERSION = "v" + __version__
AGENT = "%s/%s" % (__name__, __version__)



from setuptools import setup, find_packages
from os import path

# To use a consistent encoding
from codecs import open

HERE = path.abspath(path.dirname(__file__))

# trick to manage package versions in one place only
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
import re

VERSIONFILE = "lookmlhelper/VERSION.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    VERSIONSTRING = mo.group(1)
else:
    raise RuntimeError(
        "Unable to find version string in %s." % (VERSIONFILE, ))

# Get the long description from the README file
with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Parse requirements.txt file so to have one single source of truth
REQUIREMENTS_DATA = []
with open(path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    for l in f.readlines():
        if not l.startswith("#"):
            if ">=" in l:
                REQUIREMENTS_DATA.append([l.split(">=")[0]])
            elif "=" in l:
                REQUIREMENTS_DATA.append([l.split("=")[0]])

setup(
    name="lookml-helper",
    version=VERSIONSTRING,
    description="Package to generate LookML from GBQ databases.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/digital-science/dimensions/data-solutions/looker-lookml-generator",
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIREMENTS_DATA,
    entry_points='''
        [console_scripts]
        lookml-helper = lookmlhelper.main_cli:main_cli
        quicktest_lookmlhelper = lookmlhelper.tests.quicktest:quicktest_cli
    ''',
)






#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import os
import click

from .gbqgen.main import *

HERE = os.path.abspath(os.path.dirname(__file__))

CMD_LINE_EXAMPLES = """EXAMPLES:
$ lookmlhelper ds-data-solutions-gbq dimensions-ai.data_analytics.clinical_trials

# For more option, run from Python directly
"""




@click.command()
@click.option('--examples', is_flag=True, help='Show some examples')
@click.option(
    '--billproject',
    '-b',
    help=
    'BILLING PROJECT: the GCP billing project to access resources.'
)
@click.option(
    '--dataset',
    '-d',
    help=
    'DATASET: a fully scopes GBQ dataset eg `dimensions-ai.data_analytics.clinical_trials`.'
)
@click.pass_context
def main_cli(ctx, billproject=None, dataset=None, examples=False):
    """LOOKML-HELPER CLI.
Requires both a billproject and dataset to run. 
    """

    if examples:
        click.secho(CMD_LINE_EXAMPLES, fg="green")
        return

    if (billproject and dataset):
        run_gbq_extraction(
            gbq_billing_project=billproject,
            gbq_dataset=dataset,
        )

    else:
        click.echo(ctx.get_help())
        return


if __name__ == '__main__':
    main_cli()



#!/usr/bin/python
# -*- coding: utf-8 -*-


from string import Template
import pandas as pd
import numpy as np
import re
import os
import shutil
import pandas as pd
import itertools
from datetime import datetime

from .utils import *
from .extract_fields import *
# from ..CONFIG import *



# """## Creating the main explore

# Each field that is an ARRAY we need to unnest. In order to do this we have to create an explore with each join

# """


#
# Create for a GBQ table field that is an array a join
#

def create_base_explore_join(field):
  assert field.modus== "ARRAY" # paranoia check

  # For some fields we have offsets
  if field.nested_dimension_name in get_global_configuration()['fields_w_offset_joins']:
    offset = f" WITH OFFSET as {field.nested_dimension_name}_offset"
  else:
    offset = ""

  # Compile it all together
  array_join_lml = f"""
    {make_header(field, "join for")}
    join: {field.unnested_view_name}{{
      view_label: " {get_global_configuration()['pretty_looker_name']}"
      sql: LEFT JOIN UNNEST(${{{field.nested_view_name}.{field.nested_dimension_name}}}) as {field.unnested_view_name}{offset};;
      relationship: one_to_many
    }}"""

  return array_join_lml


#
# Create all the unnesting joins of the base explore
#

def create_all_base_explore_joins(looker_metadata, verbose=False):
  array_fields = looker_metadata[looker_metadata.modus=="ARRAY"]
  if verbose:
    printDebug("Generating joins for the ", len(array_fields), " arrays found in the table")
    printDebug(array_fields)
  return "".join(list(array_fields.apply(create_base_explore_join, axis=1)))




def write_base_explore(output_path, looker_metadata):


    base_explore_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: BASE EXPLORE
    #
    # - includes GBQ fields 1:1 from the {get_global_configuration()['pretty_looker_name']} GBQ table
    # - all arrays are unnested and JOINED
    # - no JOINs to other tables are included
    #
    ##############


    #####################
    # IMPORTS
    #####################
    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/views/unnested.view"
    #####################


    explore: {get_global_configuration()['looker_name']} {{

    label: "{get_global_configuration()['pretty_looker_name']} Basic"
    view_label: " {get_global_configuration()['pretty_looker_name']}"
    from: {get_global_configuration()['looker_name']}_unnested   # view the Explore derives from
    view_name: {get_global_configuration()['looker_name']}   # name used when referencing fields in JOINs (overrides: Explore name)

    description: "{get_global_configuration()['pretty_looker_name']} Base explore: enhanced the basic explore by extracting
    all nested information as first class objects using supplementary views."



    # ##############
    # #######
    # ## JOINS TO UNNESTED VIEWS
    # #######
    # ##############

    {create_all_base_explore_joins(looker_metadata)}

    
    # ##############
    # ## QUERIES for Quick Start analyses
    # ## https://docs.looker.com/reference/explore-params/query
    # ##############

    ##££ Add more meaningful queries here

    query: count_all {{
        label: "{get_global_configuration()['pretty_looker_name']}: count all"
        description: "Count all entries"
        measures: [{get_global_configuration()['looker_name']}.count]
        # filters: [{get_global_configuration()['looker_name']}.end_date_year: "after 0 minutes ago"]
    }}


    }}
    """


    # printDebug(base_explore_lml)
    f = open(output_path+"/explores/"+get_global_configuration()['looker_name']+".explore.lkml", "w")
    f.write(base_explore_lml)
    f.close()




# """## Creating the joined-up explore
# At this point this is sth that needs to be edited manually. But maybe we can go through the ids and other tables in the data set to suggest sth automatically
# """




def write_joined_explore(output_path, looker_metadata):

    join_explore_lml = f"""
    ##############
    ## {get_global_configuration()['capital_looker_name']}: JOINED EXPLORE
    #
    # This can be used as a blueprint/building block for joining other models to {get_global_configuration()['pretty_looker_name']}.
    #
    # Refining:
    # - import this file into your project
    # - refine the `{get_global_configuration()['looker_name']}_joined` definition
    # - includes JOINs to other tables too
    #
    # Duplicate:
    # - import `{get_global_configuration()['looker_name']}.base.explore` into your project
    # - refine it and implement your own JOINs
    # - use structure below as blueprint
    ##############


    #####################
    # IMPORTS
    #####################
    # Explores
    include: "{get_global_configuration()['looker_parent_folder']}/{get_global_configuration()['looker_name']}/explores/{get_global_configuration()['looker_name']}.explore"

    ££ CHECK IF YOU NEED MORE INCLUDES
    #####################




    explore: {get_global_configuration()['looker_name']}_joined {{

    extends: [{get_global_configuration()['looker_name']}] # explore base+unnest
    label: "{get_global_configuration()['pretty_looker_name']} Joined"

    view_label: " {get_global_configuration()['pretty_looker_name']}"
    hidden: no

    description: "{get_global_configuration()['pretty_looker_name']} Joined explore: include views for tables via joins."



    # ##############
    # #######
    # ## JOINS TO NON-{get_global_configuration()['capital_looker_name']} BASED VIEWS
    # #######
    # ##############

    # £££ take the joins in the old data table and replace nest_* with
    # #__unnested and change type from inner to left_outer
    }}
    """

    # printDebug(join_explore_lml)
    f = open(output_path+"/explores/"+get_global_configuration()['looker_name']+".joined.explore.lkml", "w")
    f.write(join_explore_lml)
    f.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-


from string import Template
import pandas as pd
import numpy as np
import re
import os
import shutil
import pandas as pd
import itertools
from datetime import datetime


try:
    from google.colab import auth
    auth.authenticate_user()
except:
    pass

from .utils import *
from .extract_fields import *
from .gen_explores import write_base_explore, write_joined_explore
from .gen_views import *
# from ..CONFIG import *


# This governs if additional information is churned out

OUTPUT_LOCAL = os.path.join(os.path.expanduser('~'), 'lookmlhelper')



def run_gbq_extraction( gbq_billing_project,
                        gbq_dataset,
                        looker_name="",
                        looker_parent_folder="",
                        fields_w_offset_joins_input="",
                        special_labels=None,
                        exclude_field_paths=None,
                        unnest_suffix="__unnested",
                        special_casts=None,
                        output_folder = None,
                        verbose=False):
    """Query a GBQ table and generates LookML docs for it.

    Parameters
    ----------
    gbq_billing_project : _type_
        _description_
    gbq_dataset : _type_
        _description_
    looker_name : str, optional
        _description_, by default ""
    looker_parent_folder : str, optional
        _description_, by default ""
    fields_w_offset_joins_input : str, optional
        _description_, by default ""
    special_labels : _type_, optional
        _description_, by default None
    exclude_field_paths : _type_, optional
        _description_, by default None
    unnest_suffix : str, optional
        _description_, by default "__unnested"
    special_casts : _type_, optional
        _description_, by default None
    output_folder : _type_, optional
        _description_, by default OUTPUT_LOCAL
    verbose : bool, optional
        _description_, by default False
    """    

    try:
        gbq_resource_name, gbq_datset_name, gbq_table_name = gbq_dataset.strip().split(".")
    except:
        printDebug("ERROR: gbq_dataset must have the format `resource.dataset.table`", "red")

    # defaults
    looker_name = looker_name or gbq_table_name
    pretty_looker_name = looker_name.title().replace("_", " ")
    looker_parent_folder = looker_parent_folder or gbq_datset_name
    if not looker_parent_folder.startswith("/"):
        looker_parent_folder = f"/{looker_parent_folder}"
    special_labels = special_labels or {}
    exclude_field_paths = exclude_field_paths or []
    special_casts = special_casts or {}
    output_folder = output_folder or OUTPUT_LOCAL

    # fill in config object
    CONFIGURATION = get_global_configuration()

    CONFIGURATION['gbq_project_id'] = gbq_billing_project
    CONFIGURATION['gbq_resource_name'] = gbq_resource_name
    CONFIGURATION['gbq_datset_name'] = gbq_datset_name
    CONFIGURATION['gbq_table_name'] = gbq_table_name
    CONFIGURATION['looker_name'] = looker_name
    CONFIGURATION['looker_parent_folder'] = looker_parent_folder
    CONFIGURATION['pretty_looker_name'] = pretty_looker_name
    CONFIGURATION['capital_looker_name'] = pretty_looker_name.upper()
    CONFIGURATION['fields_w_offset_joins'] = [x.strip() for x in fields_w_offset_joins_input.split(",")]
    CONFIGURATION['special_labels'] = special_labels
    CONFIGURATION['exclude_field_paths'] = exclude_field_paths
    CONFIGURATION['unnest_suffix'] = unnest_suffix
    CONFIGURATION['special_casts'] = special_casts
    CONFIGURATION['output_folder'] = output_folder

    printDebug(f"Querying {gbq_dataset} information schema...")

    sql = f"""
        SELECT column_name, field_path, data_type, description
        FROM `{gbq_resource_name}`.{gbq_datset_name}.INFORMATION_SCHEMA.COLUMN_FIELD_PATHS
        WHERE table_name = "{gbq_table_name}"
    """
    gbq_metadata = inspect_gbq(sql)
    gbq_metadata["modus"] = gbq_metadata.data_type.apply(get_gbq_mode)

    printDebug(f"Found {len(gbq_metadata)} fields...")

    looker_metadata = gbq_metadata.copy()

    # Exclude all field paths specified above
    for exclude_field_path in exclude_field_paths:
        looker_metadata = looker_metadata[
            ~looker_metadata.field_path.str.startswith(exclude_field_path)
        ]


    looker_metadata["array_parent_field_path"] = looker_metadata.field_path.apply(
        get_array_parent_field_path, looker_metadata=looker_metadata
    )
    looker_metadata["nested_view_name"] = looker_metadata.apply(
        get_nested_view_name, axis=1, 
    )
    looker_metadata["nested_dimension_name"] = looker_metadata.apply(
        get_nested_dimension_name, axis=1,
    )
    looker_metadata["unnested_view_name"] = looker_metadata.apply(
        get_unnested_view_name, axis=1,
    )
    looker_metadata["unnested_dimension_name"] = looker_metadata.apply(
        get_unnested_dimension_name, axis=1, 
    )
    looker_metadata["unnested_sql"] = looker_metadata.apply(get_unnested_sql, axis=1)
    looker_metadata["label"] = looker_metadata.field_path.apply(get_label)
    looker_metadata["label_plural"] = looker_metadata["label"].apply(make_label_plural)
    looker_metadata["group_label"] = looker_metadata.apply(get_group_label, axis=1, looker_metadata=looker_metadata)


    # Make nice top array headers as comments in the coding
    # This is usually if a new column_name starts but this can be sometimes the first child
    # if that field is a struct e.g. patents.orange_book
    looker_metadata["top_header"] = ""

    # Never do this but this is the easiest way to do it and we only have a few rows
    past_column_names = []
    for index, row in looker_metadata.iterrows():
        if row.modus in ["STRUCT", "ARRAY"] and row.column_name not in past_column_names:
            looker_metadata.loc[index, "top_header"] = (
                looker_metadata.loc[looker_metadata.field_path == row.column_name]
                .iloc[0]
                .label
            )
            if row.modus == "ARRAY":
                past_column_names.append(row.column_name)

    if verbose:
        printDebug(
            looker_metadata[
                [
                    "field_path",
                    "data_type",
                    "nested_view_name",
                    "nested_dimension_name",
                    "unnested_view_name",
                    "unnested_dimension_name",
                    "unnested_sql",
                    "label",
                    "label_plural",
                    "group_label",
                ]
            ]
        )

    printDebug(f"Generating LookML...")

    output_path = gen_folders(looker_name, output_folder)

    write_base_explore(output_path, looker_metadata)
    write_joined_explore(output_path, looker_metadata)
    write_raw_array_view(output_path, looker_metadata)
    write_raw_fields_view(output_path, looker_metadata)
    write_unested_array_view(output_path, looker_metadata)
    write_base_view(output_path, looker_metadata)
    write_unnested_view(output_path, looker_metadata)

    printDebug(f"Done.")